#!/usr/bin/python3
# -*- coding: utf-8 -*-
from sequence import Sequence
from math import sqrt, log
acideAminee=['A', 'R', 'N',
             'D', 'C', 'Q',
             'E', 'G', 'H',
             'I', 'L', 'K',
             'M', 'F', 'P',
             'S', 'T', 'W',
             'Y', 'V']

# Donnée prise du Site Swissprot, représente la probabilité qu'on trouve l'acide aminé
# à n'importe quelle position dans les séquences.
acideAmineePositionProbabilite= { "A" : 0.0826, "R" : 0.0553, "N" : 0.0405,
                                  "D" : 0.0546, "C" : 0.0137, "Q" : 0.0393,
                                  "E" : 0.0673, "G" : 0.0708, "H" : 0.0227,
                                  "I" : 0.0593, "L" : 0.0965, "K" : 0.0582,
                                  "M" : 0.0241, "F" : 0.0386, "P" : 0.0472,
                                  "S" : 0.0661, "T" : 0.0535, "W" : 0.0109,
                                  "Y" : 0.0292, "V" : 0.0686
                                 }
probabilite = acideAmineePositionProbabilite

class PSSM_Profil:
    def __init__(self, sequences):
        self.sequences = sequences
        self.lineSize = len(self.sequences[0])
        self.alpha = len(self.sequences)-1
        self.beta = sqrt(len(self.sequences))
        self.profilPssm = [{aa : 0 for aa in acideAminee} for _ in range(self.lineSize)]
        self.consensusSequence = ""
        self.calculNumberOcurAA()
        self.calculPseudoCounts()
        self.calculScoreAlignResidue()
        self.createConsensusSequence()

    def __len__(self):
        return len(self.profilPssm)

    def getItem(self, pos):
        return self.profilPssm[pos[0]][pos[1]]

    def getConsensusSequence(self):
        return self.consensusSequence

    def calculNumberOcurAA(self):
        for sequence in self.sequences:
            for i, aminoAcid in enumerate(sequence):
                if aminoAcid in acideAminee:
                    self.profilPssm[i][aminoAcid] += 1


    def calculPseudoCounts(self):
        for idraw in range(len(self.profilPssm)):
            for aminoAcid in self.profilPssm[idraw]:
                pseudoCounts = (self.alpha * (self.profilPssm[idraw][aminoAcid]/len(self.sequences)) + self.beta * probabilite[aminoAcid])
                pseudoCounts = pseudoCounts / (self.alpha+self.beta)
                self.profilPssm[idraw][aminoAcid] = pseudoCounts


    def calculScoreAlignResidue(self):
        for idraw in range(len(self.profilPssm)):
            for aminoAcid in self.profilPssm[idraw]:
                scoreAlignResidue = log(self.profilPssm[idraw][aminoAcid]/probabilite[aminoAcid])
                self.profilPssm[idraw][aminoAcid] = scoreAlignResidue


    def createConsensusSequence(self):
        for idraw in range(len(self.profilPssm)):
            bestScoreAlignResidue = -float("inf")
            for aminoAcid in self.profilPssm[idraw]:
                if self.profilPssm[idraw][aminoAcid] > bestScoreAlignResidue:
                    bestScoreAlignResidue = self.profilPssm[idraw][aminoAcid]
                    bestAminoAcid = aminoAcid
            self.consensusSequence += bestAminoAcid

    def __repr__(self):
        """
        Fonction qui imprime joliment la matrice, avec le nom de chaque colonne/liste
        """
        stringMatrix = str()
        stringMatrix += (" " * 8)
        for i in range(self.lineSize):
            stringMatrix += "{:^8}".format(i) #{:^8} est formatage pour avoir des colonnes/lignes identiques.
        stringMatrix += "\n"
        for aminoAcid in acideAminee:
            stringMatrix += "{:^8}".format(aminoAcid)
            for idraw in range(self.lineSize):
                stringMatrix += "{:^8}".format(round(self.profilPssm[idraw][aminoAcid], 2))
            stringMatrix += "\n"
        stringMatrix += "Séquence Consensus : {}".format(self.consensusSequence)
        return stringMatrix

    def __str__(self):
        return repr(self)

class LocalAlign():

    def __init__(self, sequence, profil, consensusSequence, gapPenalty, numberSeqToFind=1):

        self.sequence = sequence
        self.profil = profil
        self.consensusSequence = consensusSequence
        self.lineSize = len(sequence)
        self.columnSize = len(profil)
        self.gapPenalty = gapPenalty
        self.alignSequence= ""
        self.alignConsensusSequence = ""
        self.S = self.initMatrix() #matrice de Score
        self.allMaximunPos = []  # Mémoire des maximum déjà trouvé utile que pour LocalALign
        self.allVisitedPos = set()  # Mémoire de passage bactracking utile que pour LocalALign
        self.solutions = []
        self.backtrackMatrixTrace = self.initBacktrackMatrixTrace()
        self.fillMatrices()
        for k in range(numberSeqToFind):
            self.alignSequence= ""
            self.alignConsensusSequence = ""
            maxPos = self.findMaxPos()
            if maxPos == (0, 0):
                # Matrice S contient que des 0
                # -1 = valeur sentinelle pour dire plus possible de trouver des séquences.
                self.solutions.append(-1)
                break
            else:
                self.allMaximunPos.append(maxPos)
                visitedPos = list()
                self.findAlign(maxPos, visitedPos)
                self.allVisitedPos.update(visitedPos)
                self.backtrackMatrixTrace = self.initBacktrackMatrixTrace()
                self.recreateMatrix(visitedPos)

    def initMatrix(self):
        """
        Crée une matrice de taille (self.lineSize+1)x(self.columSize+1)
        """
        return [[0 for i in range(self.columnSize + 1)] for j in range(self.lineSize + 1)]


    def initBacktrackMatrixTrace(self):
        """
        Crée une matrice de direction pour savoir d'où provient
        lorsqu'on se trouve à un emplacement dans la matrice.
        Utile dans la méthode findAlign.
        """
        matrix = [["" for i in range(self.columnSize + 1)] for j in range(self.lineSize + 1)]
        for i in range(1,self.lineSize+1):
            matrix[i][0] = "T"
        for j in range(1,self.columnSize+1):
            matrix[0][j] = "L"
        return matrix


    def fillMatrices(self):
        """
        Fonction qui remplit la matrice selon des règles précices.
        Explications de remplissages voir plus bas dans le rapport.
        """
        for idraw in range(1, self.lineSize + 1):
            for idcol in range(1, self.columnSize + 1):
                tuplePos = (idraw, idcol)
                #Utile lorsque nous sommes en local pour éviter le recalcul de cellule mise à 0
                if tuplePos not in self.allVisitedPos:
                    self.S[idraw][idcol] = max(
                        self.S[idraw - 1][idcol - 1] + self.profil.getItem((idcol - 1, self.sequence[idraw - 1])),
                        self.S[idraw - 1][idcol] - self.gapPenalty,
                        self.S[idraw][idcol - 1] - self.gapPenalty)

                    if self.S[idraw][idcol] == self.S[idraw - 1][idcol - 1] + self.profil.getItem((idcol - 1, self.sequence[idraw - 1])):
                        # Moyen de se déplacer sur la diagonal
                        self.backtrackMatrixTrace[idraw][idcol] = "D"

                    elif self.S[idraw][idcol] == self.S[idraw - 1][idcol] - self.gapPenalty:
                        # Moyen de se déplacer sur le haut, donc aura un gap dans sequence consensus
                        self.backtrackMatrixTrace[idraw][idcol] = "T"

                    elif self.S[idraw][idcol] ==  self.S[idraw][idcol - 1] - self.gapPenalty:
                        # Moyen de se déplacer sur la gauche, donc aura un gap dans sequence
                        self.backtrackMatrixTrace[idraw][idcol] = "L"

                    if self.S[idraw][idcol] < 0:
                        #Utile en local car toutes valeurs négatives doivent être mise à 0.
                        self.S[idraw][idcol] = 0

    def recreateMatrix(self, visitedPos):
        """
        Fonction remettant à 0 le chemin dans la matrice  S,V,W et les recalcul arpès.
        """
        self.clearPath(visitedPos)
        self.fillMatrices()


    def clearPath(self, visitedPos):
        """
        Mets à 0 le chemin trouvé dans S,V,W
        """
        for tuplePos in visitedPos:
            self.S[tuplePos[0]][tuplePos[1]] = 0


    def findMaxPos(self):
        """
        Cherche le maximum dans la matrice score S
        """
        maximum = 0
        maxPos = (0, 0)
        for idraw in range(1, self.lineSize + 1):
            for idcol in range(1, self.columnSize + 1):
                if (idraw, idcol) not in self.allMaximunPos and (self.S[idraw][idcol] > maximum):
                    maxPos = (idraw, idcol)
                    maximum = self.S[idraw][idcol]
        return maxPos


    def findAlign(self, maxPos, visitedPos):
        idraw, idcol = maxPos[0], maxPos[1]
        while self.S[idraw][idcol] != 0:
            visitedPos.append((idraw, idcol))

            if self.backtrackMatrixTrace[idraw][idcol] == "T":
                self.alignSequence += self.sequence[idraw - 1]
                self.alignConsensusSequence += "-"
                idraw -= 1
            elif self.backtrackMatrixTrace[idraw][idcol] == "L":
                self.alignSequence += "-"
                self.alignConsensusSequence += self.consensusSequence[idcol - 1]
                idcol -= 1
            else:
                self.alignSequence += self.sequence[idraw - 1]
                self.alignConsensusSequence += self.consensusSequence[idcol - 1]
                idraw -= 1
                idcol -= 1

        self.solutions.append((self.alignSequence[::-1], self.alignConsensusSequence[::-1], self.S[maxPos[0]][maxPos[1]],
        visitedPos[-1][0]-1, visitedPos[0][0]-1, visitedPos[-1][1]-1, visitedPos[0][1]-1))

    def compareSequence(self, sequenceA, sequenceB):
        """
        Construit un string pour avoir une représentation comme sur LALIGN.
        | = match parfait entre 2 acides aminés
        : = score de substitution positif ou nul entre 2 acide aminé
        . = score de substitution négatif entre 2 acide aminé
        " " = lorsqu'il y a un gap dans une des 2 séquences.
        """
        stringRes = ""
        for i in range(len(sequenceA)):
            charA = sequenceA[i]
            charB = sequenceB[i]
            if charA == charB:
                stringRes += "|"
            elif (charA != "-"):
                stringRes += ":"
            elif (charB != "-"):
                stringRes += "."
            else:
                stringRes += " "
        return stringRes

    def __repr__(self):
        """
        Représentation des alignement trouvés.
        """
        stringToPrint = ""
        stringToPrint +="Résultat obtenue pour un alignement local entre la séquence : {} \net la séquence consensus: {} \n\n".format(
            self.sequence, self.consensusSequence)
        stringToPrint += "Linear Gap Penalty{}\n\n".format(self.gapPenalty)
        stringToPrint += "Alignements locaux obtenus :\n\n"

        for i in range(len(self.solutions)):
            if self.solutions[i] != -1:  # condtion utile que pour alignement local
                stringToPrint += "Score: {}\n\n".format(self.solutions[i][2])
                stringToPrint += "Partie de la Séquence alignée de {} à {}\nPartie de la Séquence Consensus alignée de {} à {} \n\n".format(
                self.solutions[i][3], self.solutions[i][4], self.solutions[i][5], self.solutions[i][6])
                for j in range(0, len(self.solutions[i][0]), 60):
                    stringToPrint += "Séquence alignée:  {}\n{}{}\nConsensus alignée: {}\n\n".format(self.solutions[i][0][j:j + 60], " " * 19,
                                                                      self.compareSequence(
                                                                          self.solutions[i][0][j:j + 60],
                                                                          self.solutions[i][1][j:j + 60]),
                                                                      self.solutions[i][1][j:j + 60])
            else:
                stringToPrint += "Impossible de trouver plus de séquences alignées\n"
        return stringToPrint

    def __str__(self):
        return repr(self)


def main():
    clustalSequences = [sequence.getSequence() for sequence in Sequence.loadData('./msaresults-CLUSTAL.fasta')]
    profilClustal = PSSM_Profil(clustalSequences)
    muscleSequences = [sequence.getSequence() for sequence in Sequence.loadData('./msaresults-MUSCLE.fasta')]
    print(PSSM_Profil(muscleSequences))
    # proteinSequence = [sequence.getSequence() for sequence in Sequence.loadData('./protein-sequences.fasta')]
    # for sequence in proteinSequence:
    #     testlocal = LocalAlign( sequence, profilClustal, profilClustal.getConsensusSequence(), gapPenalty = 9, numberSeqToFind=4)
    #     print(testlocal)
if __name__ == '__main__':
    main()
    # clustalSequences = [sequence.getSequence() for sequence in Sequence.loadData('./msaresults-CLUSTAL.fasta')]
    # muscleSequences = [sequence.getSequence() for sequence in Sequence.loadData('./msaresults-MUSCLE.fasta')]
    # profilClustal = PSSM_Profil(clustalSequences)
    # proteinSequence = [sequence.getSequence() for sequence in Sequence.loadData('./protein-sequences.fasta')]
    # for sequence in proteinSequence:
    #     testlocal = LocalAlign( sequence, profilClustal, profilClustal.getConsensusSequence(), gapPenalty = 9, numberSeqToFind=4)
    #     print(testlocal)
    # for sequence in proteinSequence:
    #     testlocal = LocalAlign( sequence, profilClustal, profilClustal.getConsensusSequence(), gapPenalty = 1, numberSeqToFind=2)
    #     print(testlocal)
