#!/usr/bin/python3

from math import log

import matplotlib.pyplot as matplot
import numpy as np
import pandas as pd
from sequence import Sequence

# Acide aminé valide à utiliser
aminoAcids = ['A', 'R', 'N',
              'D', 'C', 'Q',
              'E', 'G', 'H',
              'I', 'L', 'K',
              'M', 'F', 'P',
              'S', 'T', 'W',
              'Y', 'V']
# Différentes structures possibles
structures = ["H", "E", "C"]
#Structure n'étant pas la structure S
nStructures = {s: [s2 for s2 in structures if s2 != s] for s in structures}
#Table de correspondance pour les différentes structures
symbolsTable = {"H": "H",
                "G": "H",
                "I": "H",
                "E": "E",
                "B": "E",
                "T": "C",
                "C": "C",
                "S": "C",
                " ": "C"}

windowSize = 8  # Taille de la fenêtre pour la gauche et la droite
# Phrase utilisé lors du l'affichage des graphiques
plotLabelString = {"Q3": ["Moyenne des score Q3", "Scores Q3", "Score Q3", "Nombre total de Q3"],
                   ("MCC"): ["Moyenne des score MCC", "Scores MCC de la structure {}", "Score MCC",
                             "Nombre total de MCC"],
                   ("Sens"): ["Moyenne du taux de sensibilité", "Taux de sensibilités de la structure {}",
                              "Taux de sensibilité", "Nombre total de taux de sensibilité"],
                   ("Spec"): ["Moyenne du taux de spécificité", "Taux de spécificités de la structure {}",
                              "Taux de spécificité", "Nombre total de taux de spécificité"],
                   }
# Couleurs utilisés pour les différents graphiques
colorPlot = {"": "blue", "H": "purple", "E": "yellow", "C": "green"}


class Gor():

    def __init__(self):
        """
        Fonction définissant les variables sans les initialisées de manière spécifique
        """
        # Variables spécifiques au training du predicteur, dont certaines seront utilisées lors du testing du prédicteur
        # liste contenant des sequences et structures des séquences de protéines utilisées pour le Training
        self.__sequencesAndStructuresTraining = list()
        # dictionnaire de fréquences d'observer la structure S dans toutes les séquences,
        # ayant comme clef une structure et comme valeur le nombre d'observation
        self.__fS = dict()
        # dictionnaire de fréquences d'observer la structure S avec un acide aminé R dans toutes les séquences,
        # ayant comme clef une structureS, acide aminé R et comme valeur le nombre d'observation.
        self.__fSR = dict()
        # dictionnaire de fréquences d'observer la structure S avec 2 acides aminés Rj et Rj+m dans toutes les séquences,
        # ayant comme clef une structureS, acide aminé R et comme valeur le nombre d'observation.
        self.__fSRRm = dict()
        # dictionnaire de fréquences d'observer la structure non S dans toutes les séquences,
        # ayant comme clef une structure et comme valeur le nombre d'observation
        self.__fnS = dict()
        # dictionnaire de fréquences d'observer la structure non S avec un acide aminé R dans toutes les séquences,
        # ayant comme clef une structureS, acide aminé R et comme valeur le nombre d'observation.
        self.__fnSR = dict()
        # dictionnaire de fréquences d'observer la structure non S avec 2 acides aminés Rj et Rj+m dans toutes les séquences,
        # ayant comme clef une structureS, acide aminé R et comme valeur le nombre d'observation.
        self.__fnSRRm = dict()
        # dictionnaire de score/probabilité d'observer un acide aminé R dans une structure S,
        # ayant comme clef une structure S, acide aminé R et un score/probabilité comme valeur
        self.__scoreSR = dict()
        # dictionnaire de score/probabilité d'observer 2 acides aminés Rj,Rj+m dans une structure S,
        # ayant comme clef une structure S, acide aminé Rj,Rj+m et un score/probabilité comme valeur
        self.__scoreSRRm = dict()

        # Variables spécifiques au testing du predicteur
        # liste contenant des sequences, structures et identifiant des séquences de protéines utilisées pour le Training
        self.__sequencesAndStructuresTesting = list()
        # dictionnaire des prédictions obtenues avec comme clef l'identifiant de la séquence utilisé pour la prédictions
        # et comme valeur la prédictions de la structure.
        self.__resultsPredictions = dict()
        # dictionnaire des scores Q3,MCC, Sensibilité, Spécificité obtenues avec comme clef l'identifiant
        # de la séquence/structure et comme valeur le score Q3,MCC, Sensibilité, Spécificité de la structure,
        # a aussi une clef "avg" et "stdDev" pour la moyenne et l'écart-type.
        self.__resultsQ3 = dict()
        self.__resultsMCC = dict()
        self.__resultsSensitivity = dict()
        self.__resultsSpecificity = dict()

    def initAndRunTraining(self, proteinSequencesTraining):
        """
        Fonctiond qui initialise les variables utiles au training du dataset grâce aux donnnées transmissent
        par le paramètre "proteinSequencesTraining" et lance les différentes méthodes effectuant le training
        de notre prédicteur.

        :param proteinSequencesTraining: liste contenant les intances de la Classe Sequence contenant les données utiles
                                         à l'entrainement.
        """
        self.__sequencesAndStructuresTraining = [
            (proteinSequenceTraining.getSequence(), proteinSequenceTraining.getStructure()) for proteinSequenceTraining
            in proteinSequencesTraining]
        self.__fS = {symbols: 0 for symbols in structures}
        self.__fSR = {(symbols, aminoAcid): 0 for aminoAcid in aminoAcids for symbols in structures}
        self.__fSRRm = {(symbols, aminoAcid, aminoAcidNeighbour): 0 for aminoAcid in aminoAcids for aminoAcidNeighbour
                        in
                        aminoAcids for symbols in structures}
        self.__fnS = {symbols: 0 for symbols in structures}
        self.__fnSR = {(symbols, aminoAcid): 0 for aminoAcid in aminoAcids for symbols in structures}
        self.__fnSRRm = {(symbols, aminoAcid, aminoAcidNeighbour): 0 for aminoAcid in aminoAcids for aminoAcidNeighbour
                         in
                         aminoAcids for symbols in structures}
        self.__scoreSR = {(s, aa): 0 for aa in aminoAcids for s in structures}
        self.__scoreSRRm = {(symbols, aminoAcid, aminoAcidNeighbour): 0 for aminoAcid in aminoAcids for
                            aminoAcidNeighbour
                            in aminoAcids for symbols in structures}
        self.__calculFrequencies()
        self.__calculScore()

    def initTestingVariable(self, proteinSequencesTesting):
        """
        Fonction qui initialise les variables utiles au bon déroulement de notre prédicteur grâce aux données contenues
        dans le paramètre "proteinSequencesTesting".

        :param proteinSequencesTesting: liste contenant les intances de la Classe Sequence contenant les données utiles
                                         au testing du prédicteur.
        """
        self.__sequencesAndStructuresTesting = [
            (proteinSequence.getSequence(), proteinSequence.getStructure(), proteinSequence.getIdentificator()) for
            proteinSequence in proteinSequencesTesting]
        self.__resultsPredictions = {item[2]: "" for item in self.__sequencesAndStructuresTesting}
        self.__resultsQ3 = {item[2]: 0 for item in self.__sequencesAndStructuresTesting}
        self.__resultsMCC = {item[2]: 0 for item in self.__sequencesAndStructuresTesting}
        self.__resultsSensitivity = {item[2]: 0 for item in self.__sequencesAndStructuresTesting}
        self.__resultsSpecificity = {item[2]: 0 for item in self.__sequencesAndStructuresTesting}

    def __calculFrequencies(self):
        """
        Fonction calculant les différents fréquences:
            - observer la structure S
            - observer un acide aminé R dans une structure S
            - observer 2 acides aminées Rj,Rj+m dans une structure S
        Et aussi calcul l'inverse de ces fréquences dans une structure non S.
        Tous cela fait pour toutes les séquences.
        """
        # Calcul des fréquences pour les structures S
        for sequence, structure in self.__sequencesAndStructuresTraining:
            for index in range(len(sequence)):
                aminoAcid = sequence[index]
                symnolsStructure = structure[index]
                self.__fS[symnolsStructure] += 1
                self.__fSR[symnolsStructure, aminoAcid] += 1
                sequenceNeighbour = sequence[max(0, index - windowSize):index] + sequence[
                                                                                 min(index + 1, len(sequence)): min(
                                                                                     index + 1 + windowSize,
                                                                                     len(sequence))]
                for neighbourAminoAcid in sequenceNeighbour:
                    self.__fSRRm[(symnolsStructure, aminoAcid, neighbourAminoAcid)] += 1

        #Calcul des fréquences pour les structures non S
        for s in nStructures:
            for s2 in nStructures[s]:
                self.__fnS[s] += self.__fS[s2]
                for aminoAcid in aminoAcids:
                    self.__fnSR[(s, aminoAcid)] += self.__fSR[(s2, aminoAcid)]
                    for aa_v in aminoAcids:
                        self.__fnSRRm[(s, aminoAcid, aa_v)] += self.__fSRRm[(s2, aminoAcid, aa_v)]

    def __calculScore(self):
        """
        Fonction calculant les score/problitite:
            - score/probabilité d'observé l'acide aminé R dans une structure S
            - score/probabilité d'observé 2 acides aminée Rj,Rj+m dans une structure S
        Score/probabilité calculé à partir de formules et des fréquences calculées.
        """
        for s in structures:
            for aminoAcid in aminoAcids:
                self.__scoreSR[(s, aminoAcid)] += log(self.__fSR[(s, aminoAcid)] / self.__fnSR[(s, aminoAcid)])
                self.__scoreSR[(s, aminoAcid)] += log(self.__fnS[s] / self.__fS[s])
                for aa_v in aminoAcids:
                    self.__scoreSRRm[(s, aminoAcid, aa_v)] += log(
                        self.__fSRRm[(s, aminoAcid, aa_v)] / self.__fnSRRm[(s, aminoAcid, aa_v)])
                    self.__scoreSRRm[(s, aminoAcid, aa_v)] += log(
                        self.__fnSR[(s, aminoAcid)] / self.__fSR[(s, aminoAcid)])

    def computeAllPredictedStructure(self):
        """
        Fonction qui prédite les structures de tout les séquences utilisé pour le testing
        et les stocke dans l'attribut "__resultsPredictions".
        """
        for item in self.__sequencesAndStructuresTesting:
            sequence = item[0]
            structure = item[1]
            proteinIdentificator = item[2]
            self.__resultsPredictions[proteinIdentificator] = self.__findPredictedStructure(sequence)

    def __findPredictedStructure(self, sequence):
        """
        Fonction qui prédit la structure d'une séquence d'acide aminé, en se basant sur les différents scores calculés.
        :param sequence: séquence d'acide aminé d'une protéine dont on veut en prédire la structure.
        :return: un string "structurePrediction" qui contient la séquence prédite.
        """
        structurePrediction = ""
        for index in range(len(sequence)):
            scoreStructures = {s: 0 for s in structures}
            currentAminoAcid = sequence[index]
            for structure in structures:
                scoreStructures[structure] += self.__scoreSR[(structure, currentAminoAcid)]
                sequenceNeighbour = sequence[max(0, index - windowSize):index] + sequence[min(index + 1, len(
                    sequence)): min(index + 1 + windowSize, len(sequence))]
                for neighbourAminoAcid in sequenceNeighbour:
                    scoreStructures[structure] += self.__scoreSRRm[(structure, currentAminoAcid, neighbourAminoAcid)]
            structurePrediction += max(scoreStructures, key=scoreStructures.get)
        return structurePrediction

    def __countSameStructure(self, realStructure, predictedStructure):
        """
        Fonction "helper" pour compter le nombre de fois qu'une prédiction est correct
        par rapport à la structure de base(réel).
        :param realStructure: structure de base réel correspondant d'une séquence d'acide aminé
        :param predictedStructure: structure prédite sur base des différents scores et sur une séquence d'acide aminé
        :return: le nombre de prédiction correcte.
        """
        correctStructure = 0
        for index in range(len(realStructure)):
            if realStructure[index] == predictedStructure[index]:
                correctStructure += 1
        return correctStructure

    def computeQ3(self):
        """
        Fonction calculant le score Q3 de chaque structure prédite d'une séquence d'acide aminé
        comparé à la structure réel de cette séquence d'acide aminé, la moyenne des scores Q3 et leurs écart-type.
        """
        # Calcul des scores Q3
        for item in self.__sequencesAndStructuresTesting:
            structure = item[1]
            proteinIdentificator = item[2]
            self.__resultsQ3[proteinIdentificator] = self.__computeIndividualQ3(structure,
                                                                                self.__resultsPredictions[
                                                                                    proteinIdentificator])
        listQ3 = list(self.__resultsQ3.values())
        # calcul de la moyenne
        meanQ3 = np.mean(listQ3)
        self.__resultsQ3["avg"] = meanQ3
        # calcul de l'écart-type
        stdDev = np.std(listQ3)
        self.__resultsQ3["stdDev"] = stdDev

    def __computeIndividualQ3(self, realStructure, predictedStructure):
        """
        Fonction calculant le score Q3 de chaque structure réel et prédite correspondant
        à une séquence d'acide aminé individuellement
        :param realStructure: structure de base réel correspondant d'une séquence d'acide aminé
        :param predictedStructure: structure prédite sur base des différents scores et sur une séquence d'acide aminé
        :return: un nombre flottant représant le score Q3
        """
        correctStructure = self.__countSameStructure(realStructure, predictedStructure)
        resultScore = correctStructure / len(realStructure)
        return resultScore

    def __calcul_TN_FN_TP_FP(self, realStructure, predictedStructure):
        """
        Fonction calculant les mesures statistiques TN, TP, FP, FN pour chaque type de structure
        qui aident au calcul de la qualité des prédictions de score MCC,
        se basant sur la structure prédite et la structure réel associé à une séquence d'acide aminé
        :param realStructure: structure de base réel correspondant d'une séquence d'acide aminé
        :param predictedStructure: structure prédite sur base des différents scores et sur une séquence d'acide aminé
        :return: une liste de mesures statistiques pour les différentes types de structures.
        """
        TN_H, FN_H, FP_H, TP_H = 0, 0, 0, 0
        TN_E, FN_E, FP_E, TP_E = 0, 0, 0, 0
        TN_C, FN_C, FP_C, TP_C = 0, 0, 0, 0

        for index in range(len(realStructure)):
            if predictedStructure[index] == "H":
                if realStructure[index] == "C":
                    FP_H += 1
                    FN_C += 1
                    TN_E += 1
                elif realStructure[index] == "H":
                    TP_H += 1
                    TN_C += 1
                    TN_E += 1
                elif realStructure[index] == "E":
                    FP_H += 1
                    TN_C += 1
                    FN_E += 1
            elif predictedStructure[index] == "C":
                if realStructure[index] == "C":
                    TP_C += 1
                    TN_H += 1
                    TN_E += 1
                elif realStructure[index] == "H":
                    FP_C += 1
                    FN_H += 1
                    TN_E += 1
                elif realStructure[index] == "E":
                    FP_C += 1
                    FN_E += 1
                    TN_H += 1
            elif predictedStructure[index] == "E":
                if realStructure[index] == "C":
                    FP_E += 1
                    FN_C += 1
                    TN_H += 1
                elif realStructure[index] == "H":
                    FP_E += 1
                    FN_H += 1
                    TN_C += 1
                elif realStructure[index] == "E":
                    TP_E += 1
                    TN_C += 1
                    TN_H += 1

        return TP_H, TP_C, TP_E, TN_H, TN_C, TN_E, FP_H, FP_C, FP_E, FN_H, FN_C, FN_E

    def computeMCC_Sensitivity_Specificity(self):
        """
        Fonction calculant l'ensemble des valeurs pouvant attesté de la qualité du prédicteurs pour toutes les séquences:
            - Score MCC
            - Taux de sensibilité
            - Taux de spécificité
        Pour chaque type de score/taux, calcul des moyenne et écar-type inclus.
        """
        for item in self.__sequencesAndStructuresTesting:
            structure = item[1]
            proteinIdentificator = item[2]
            result = self.__computeIndividualMCC_Sensitivity_Specificity(structure,
                                                                         self.__resultsPredictions[
                                                                             proteinIdentificator])
            self.__resultsMCC[proteinIdentificator] = result[0]
            self.__resultsSensitivity[proteinIdentificator] = result[1]
            self.__resultsSpecificity[proteinIdentificator] = result[2]

        listMCC = [[], [], []]
        listSensitivity = [[], [], []]
        listSpecificity = [[], [], []]
        for value in zip(self.__resultsMCC.values(), self.__resultsSensitivity.values(),
                         self.__resultsSpecificity.values()):
            for i in range(3):
                listMCC[i].append(value[0][i])
                listSensitivity[i].append(value[1][i])
                listSpecificity[i].append(value[2][i])
        self.__resultsMCC["avg"] = []
        self.__resultsMCC["stdDev"] = []
        self.__resultsSpecificity["avg"] = []
        self.__resultsSpecificity["stdDev"] = []
        self.__resultsSensitivity["avg"] = []
        self.__resultsSensitivity["stdDev"] = []
        for i in range(3):
            self.__resultsMCC["avg"].append(np.mean(listMCC[i]))
            self.__resultsMCC["stdDev"].append(np.std(listMCC[i]))
            self.__resultsSpecificity["avg"].append(np.mean(listSpecificity[i]))
            self.__resultsSpecificity["stdDev"].append(np.std(listSpecificity[i]))
            self.__resultsSensitivity["avg"].append(np.mean(listSensitivity[i]))
            self.__resultsSensitivity["stdDev"].append(np.std(listSensitivity[i]))

    def __calculScoreMCC(self, TP, TN, FP, FN):
        """
        Calcul du score MCC sur base des mesures statistiques données en paramètre.
        :param TP: mesure statistique représentant le nombre de prédiction correctement indentifié
        :param TN: mesure statistique représentant le nombre de prédiction correctement rejeté
        :param FP: mesure statistique représentant le nombre de prédiction incorrectement indentifié
        :param FN: mesure statistique représentant le nombre de prédiction incorrectement rejeté
        :return: le score MCC ou 0 0 si le dénominateur est nul.
        """
        scoreMCC = 0
        if (TP + FP) * (TP + FN) * (TN + FP) * (TN + FN) != 0:
            scoreMCC = ((TP * TN) - (FP * FN)) / (((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)) ** (1 / 2))
        return scoreMCC

    def __calculSensRate(self, TP, TN, FP, FN):
        """
        Calcul du taux de sensibilité sur base des mesures statistiques données en paramètre.
        :param TP: mesure statistique représentant le nombre de prédiction correctement indentifié
        :param TN: mesure statistique représentant le nombre de prédiction correctement rejeté
        :param FP: mesure statistique représentant le nombre de prédiction incorrectement indentifié
        :param FN: mesure statistique représentant le nombre de prédiction incorrectement rejeté
        :return: le taux de sensibilité ou 0 si le dénominateur est nul.
        """
        rateSens = 0
        if (TP + FN) != 0:
            rateSens = TP / ((TP + FN))
        return rateSens

    def __calculSpecRate(self, TP, TN, FP, FN):
        """
        Calcul du taux de spécificité sur base des mesures statistiques données en paramètre.
        :param TP: mesure statistique représentant le nombre de prédiction correctement indentifié
        :param TN: mesure statistique représentant le nombre de prédiction correctement rejeté
        :param FP: mesure statistique représentant le nombre de prédiction incorrectement indentifié
        :param FN: mesure statistique représentant le nombre de prédiction incorrectement rejeté
        :return: le taux de spécificité ou 0 si le dénominateur est nul.
        """
        rateSpec = 0
        if (TN + FP) != 0:
            rateSpec = TN / ((TN + FP))
        return rateSpec

    def __computeIndividualMCC_Sensitivity_Specificity(self, realStructure, predictedStructure):
        """
        Fonction calculant les 3 types de score/taux de chaque type structure
        sur base de la structure prédite et structure réel se rapportant à une séquence d'acide aminé
        :param realStructure: structure de base réel correspondant d'une séquence d'acide aminé
        :param predictedStructure: structure prédite sur base des différents scores et sur une séquence d'acide aminé
        :return: les 3 types de score/taux de chaque type de structure
        """
        TP_H, TP_C, TP_E, TN_H, TN_C, TN_E, FP_H, FP_C, FP_E, FN_H, FN_C, FN_E = self.__calcul_TN_FN_TP_FP(
            realStructure, predictedStructure)

        scoreMCC_H = self.__calculScoreMCC(TP_H, TN_H, FP_H, FN_H)
        scoreMCC_E = self.__calculScoreMCC(TP_E, TN_E, FP_E, FN_E)
        scoreMCC_C = self.__calculScoreMCC(TP_C, TN_C, FP_C, FN_C)

        sens_H = self.__calculSensRate(TP_H, TN_H, FP_H, FN_H)
        sens_E = self.__calculSensRate(TP_E, TN_E, FP_E, FN_E)
        sens_C = self.__calculSensRate(TP_C, TN_C, FP_C, FN_C)

        spec_H = self.__calculSpecRate(TP_H, TN_H, FP_H, FN_H)
        spec_E = self.__calculSpecRate(TP_E, TN_E, FP_E, FN_E)
        spec_C = self.__calculSpecRate(TP_C, TN_C, FP_C, FN_C)

        return [scoreMCC_H, scoreMCC_E, scoreMCC_C], [sens_H, sens_E, sens_C], [spec_H, spec_E, spec_C]

    def getScoreQ3(self):
        """
        Fonction renvoyant le dictionnaire de score Q3
        """
        return self.__resultsQ3

    def getScoreMCC(self):
        """
        Fonction renvoyant le dictionnaire de score MCC
        """
        return self.__resultsMCC

    def getPrediction(self):
        """
        Fonction renvoyant le dictionnaire contenant toutes les structures prédites
        """
        return self.__resultsPredictions

    def getFrequencies(self):
        """
        Fonction renvoyant les dictionnaires de fréquences
        """
        return self.__fS, self.__fSR, self.__fSRRm, self.__fnS, self.__fnSR, self.__fnSRRm

    def getScores(self):
        """
        Fonction renvoyant les scores/probabilités
        """
        return self.__scoreSR, self.__scoreSRRm

    def getSensitivity(self):
        """
        Fonction renvoyant le dictionnaire du taux de sensibilité
        """
        return self.__resultsSensitivity

    def getSpecificity(self):
        """
        Fonction renvoyant le dictionnaire du taux de spécificité
        """
        return self.__resultsSpecificity


def main(trainingFile, testingFile, fullDisplayDetails=False):
    """
    Fonction principale faisant appel à la classe Gor pour effectuer le training et testing, et imprime les résultats.
    :param trainingFile: fichier .fasta utilisé pour le training du prédicteur
    :param testingFile: fichier .fasta utilisé pour le testing du prédicteur
    :param fullDisplayDetails: valeur booléen pour l'affiche de résultats plus en détails(résultat de chaque séquence)
    que des résultats généraux.
    """
    proteinSequencesTraining, proteinSequencesTesting = Sequence.loadData(trainingFile), Sequence.loadData(testingFile)
    realProteinStructure = {item.getIdentificator(): item.getStructure() for item in proteinSequencesTesting}
    gor = Gor()
    gor.initAndRunTraining(proteinSequencesTraining)
    gor.initTestingVariable(proteinSequencesTesting)
    gor.computeAllPredictedStructure()
    gor.computeQ3()
    gor.computeMCC_Sensitivity_Specificity()
    scoreSR, scoreSRRm = gor.getScores()
    dataToDisplay = [gor.getScoreMCC(), gor.getScoreQ3(), gor.getSensitivity(), gor.getSpecificity(), scoreSR, scoreSRRm]
    if fullDisplayDetails:
        dataToDisplay.append(gor.getPrediction())
        dataToDisplay.append(realProteinStructure)
    printResults(dataToDisplay, fullDisplayDetails)


def colorDifference(realStructure, predictedStructure):
    """
    Fonction mettant en couleur les différence et correspondance dans la structure prédite et la structure réel.
    :param realStructure: structure de base réel correspondant d'une séquence d'acide aminé
    :param predictedStructure: structure prédite sur base des différents scores et sur une séquence d'acide aminé
    :return: 2 string contenant la structure prédite en couleur pour l'un et pour l'autre, la structure réel en couleur
    """
    colorPredictedStructure = ""
    colorRealStructure = ""
    for i in range(len(realStructure)):
        if realStructure[i] != predictedStructure[i]:
            CSI = "\x1B["
            colorLetterPredicted = CSI + str(31) + "m" + predictedStructure[i] + CSI + "0m"
            colorLetterReal = CSI + str(31) + "m" + realStructure[i] + CSI + "0m"

        else:
            CSI = "\x1B["
            colorLetterPredicted = CSI + str(32) + "m" + predictedStructure[i] + CSI + "0m"
            colorLetterReal = CSI + str(32) + "m" + realStructure[i] + CSI + "0m"

        colorPredictedStructure += colorLetterPredicted
        colorRealStructure += colorLetterReal
    return colorRealStructure, colorPredictedStructure


def printScore(scoreSR, scoreSRRm):
    """
    Fonction qui affiche des tableaux avec les scores/probabilités:
        - d'observer un acide aminé R dans une structure S (scoreSR)
        - d'observer un acide aminé Rj,Rj+m dans une structure S (scoreSRRm)
    Affichage réalisé à l'aide de la librairie "pandas"
    """
    scoreSR, scoreSRRm = dictToMatrix(scoreSR, scoreSRRm)
    print("\nVoici le tableau des scores/probalitées (propensity) d'observer un acide aminé R dans une structure S:")
    print("")
    pd.set_option('display.height', 115)
    pd.set_option('display.max_rows', 25)
    pd.set_option('display.max_columns', 25)
    pd.set_option('display.width', 115)
    table = pd.DataFrame(scoreSR)
    table.columns = aminoAcids
    table.index = structures
    print(table)
    print("")
    print(
        "Voici les différents tableaux de scores/probailitées (propensity) d'observer deux acides aminés Rj et Rj+m\ndans une structure S:\n")
    for index, s in enumerate(structures):
        print("Tableau de score/probabilité d'observer deux acides aminés Rj et Rj+m dans une structure {}".format(s))
        print("")
        table = pd.DataFrame(scoreSRRm[index])
        table.columns = aminoAcids
        table.index = aminoAcids
        print(table)
        print()


def printResults(dataToDisplay, fullDisplayDetails):
    """
    Fonction affichant les différents résultats et faisant appel à d'autre fonctions d'affichage.
    :param dataToDisplay: données à afficher
    :param fullDisplayDetails: valeur booléen pour l'affiche de résultats plus en détails(résultat de chaque séquence)
    """
    print("-------Résultat de l'algorithme GorIII-------\n\n")
    print(
        "Entrainement réalisé avec 3099 séquences qui représentent les 3000 premiers fichiers dssp représentant\nnotre training dataset, dont les séquences sont contenues dans dataFromDsspTraining.fasta.")
    printScore(dataToDisplay[4], dataToDisplay[5])
    if fullDisplayDetails:
        print(
            "Les résultats suivant ont été ablis avec le reste des fichiers dssp(données) du répertoire dssp_test,\ndont les séquences sont contenues dans le fichier dataFromDssp_testTesting.fasta, 724 séquences pour 713 fichiers.")
        print("\nRésultat individuel de chaque séquence:")
        for key in dataToDisplay[7]:
            print("Identifiant séquence: {}".format(key))
            print("Score Q3: {}\n".format(dataToDisplay[1][key]))
            for i in range(3):
                print("Score MCC pour la structure {}: {}".format(structures[i], dataToDisplay[0][key][i]))
                print("Taux de sensibilité pour la structure {}: {}".format(structures[i], dataToDisplay[2][key][i]))
                print("Taux de spécificité pour la structure {}: {}".format(structures[i], dataToDisplay[3][key][i]))
            for j in range(0, len(dataToDisplay[7][key]), 80):
                colorRealStructure, colorPredictedStructure = colorDifference(dataToDisplay[7][key][j:j + 80],
                                                                              dataToDisplay[6][key][j:j + 80])
                print("Structure réel:    {}".format(colorRealStructure))
                print("Structure prédite: {}".format(colorPredictedStructure))
            print("")
    else:
        print(
            "Les résultats suivant ont été établis avec le reste des fichiers dssp(données) du répertoire dssp utilisé\nlors du training de notre prédicteur et dont les séquences sont contenues dans le fichier dataFromDsspTesting.fasta, \n724 séquences pour 713 fichiers.")

    print("\nRésultats généraux pour l'ensemble des séquences:")
    print("")
    print("Moyenne des scores Q3 (basé sur chaque Q3 individuel): {}".format(dataToDisplay[1]["avg"]))
    print("Écart-type des scores Q3: {}".format(format(dataToDisplay[1]["stdDev"])))
    listQ3 = [dataToDisplay[1][key] for key in dataToDisplay[1] if key not in ["avg", "stdDev"]]
    plotHistogram(listQ3, dataToDisplay[1]["avg"], "Q3")
    listMCC, listSens, listSpec = dictToList(dataToDisplay[0], dataToDisplay[2], dataToDisplay[3])
    for i in range(3):
        print("Score MCC moyen pour la structure {}(basé sur chaque MCC individuel): {}".format(structures[i],
                                                                                                 dataToDisplay[0][
                                                                                                     "avg"][i]))
        print("Écart-type des scores MCC pour la structure {}: {}".format(structures[i],
                                                                          format(dataToDisplay[0]["stdDev"][i])))
        plotHistogram(listMCC[i], dataToDisplay[0]["avg"][i], "MCC", structures[i])
        print(
            "Moyenne du taux de sensibilité pour la structure {}: {}".format(structures[i], dataToDisplay[2]["avg"][i]))
        print("Écart-type du taux de sensibilité pour la structure {}: {}".format(structures[i],
                                                                                  dataToDisplay[2]["stdDev"][i]))
        plotHistogram(listSens[i], dataToDisplay[2]["avg"][i], "Sens", structures[i])
        print(
            "Moyenne du taux de spécificité pour la structure {}: {}".format(structures[i], dataToDisplay[3]["avg"][i]))
        print("Écart-type du taux de spécificité pour la structure {}: {}".format(structures[i],
                                                                                  dataToDisplay[3]["stdDev"][i]))
        plotHistogram(listSpec[i], dataToDisplay[3]["avg"][i], "Spec", structures[i])
    print("")


def plotHistogram(dataToPlot, mean, keyLabel, structure=""):
    """
    Fonction qui crée et affiche un histogramme des données donnés en paramètre(dataToplot)
    :param dataToPlot: données à afficher
    :param mean: moyenne des données
    :param keyLabel: string servant pour l'affichage des bons string
    :param structure: utile pour  les Mcc, Sensibilité et Spécificité
    """
    matplot.figure(figsize=(12, 12))
    matplot.axvline(mean, color='r', linestyle='dashed', linewidth=2, label=plotLabelString[keyLabel][0])
    matplot.hist(dataToPlot, bins=150, edgecolor='black', color=colorPlot[structure])
    matplot.title(plotLabelString[keyLabel][1].format(structure))
    matplot.xlabel(plotLabelString[keyLabel][2])
    matplot.ylabel(plotLabelString[keyLabel][3])
    matplot.legend()
    matplot.show()


def dictToMatrix(dictSR, dictSRRm):
    """
    Fonction qui transforme les dictionnaires en listes, car plus facile pour l'affichage des données avec "pandas".
    :param dictSR: score d'observer un acide aminé R dans une structure S
    :param dictSRRm: score d'observer un acide aminé Rj,Rj+m dans une structure S
    :return:
    """
    matrixSR = [[0 for i in range(len(aminoAcids))] for _ in range(len(structures))]
    matrixSRRm = [[[0 for j in range(len(aminoAcids))] for i in range(len(aminoAcids))] for _ in range(len(structures))]
    for key in dictSR:
        s, aa = key[0], key[1]
        index_s, index_aa = structures.index(s), aminoAcids.index(aa)
        matrixSR[index_s][index_aa] = round(dictSR[key], 3)
    for key in dictSRRm:
        s, aa, aa_v = key[0], key[1], key[2]
        index_s, index_aa, index_aa_v = structures.index(s), aminoAcids.index(aa), aminoAcids.index(aa_v)
        matrixSRRm[index_s][index_aa][index_aa_v] = round(dictSRRm[s, aa, aa_v], 3)
    return matrixSR, matrixSRRm


def dictToList(dictMCC, dictSens, dictSpec):
    """
    Fonction transformant les dictionnaires de score MCC, taux de sensibilité et de spécificité µ
    en listes contenant une liste pour chaque type de structure en excluant la moyenne et l'écart-type.
    """
    listMCC = [[], [], []]
    listSens = [[], [], []]
    listSpec = [[], [], []]
    for key in dictMCC:
        if key not in ["avg", "stdDev"]:
            for i in range(3):
                listMCC[i].append(dictMCC[key][i])
                listSens[i].append(dictSens[key][i])
                listSpec[i].append(dictSpec[key][i])

    return listMCC, listSens, listSpec

if __name__ == '__main__':
    main("dataFromDsspTraining.fasta", "dataFromDsspTesting.fasta")
    # main("dataFromDsspTraining.fasta", "dataFromDssp_testTesting.fasta", True)
