#!/usr/bin/python3
from collections import namedtuple

# Acide aminé valide à utiliser
aminoAcids = ['A', 'R', 'N',
              'D', 'C', 'Q',
              'E', 'G', 'H',
              'I', 'L', 'K',
              'M', 'F', 'P',
              'S', 'T', 'W',
              'Y', 'V']
structures = ["H", "E", "C"]
symbolsTable = {"H": "H",
                "G": "H",
                "I": "H",
                "E": "E",
                "B": "E",
                "T": "C",
                "C": "C",
                "S": "C",
                " ": "C"}


def parserCathFile(fileName, trainingNumber):
    """
    Fonction effectuant le parsing du fichier fileName.
    On retire chaque ligne avec le nom du fichier et le caractère à considérer et stocké dans resultFileParsing
    chaque donnée des lignes qu'on retire est stocké dans un namedtuple dsspToUse("file","charchain").

    :param trainingNumber
    :param fileName: fichier à parser soit CATH_info ou CATH_info_test
    :return: resultFileParsing: liste avec toutes les ligne du fichiers fileName.
    """
    resultFileParsingTraining, resultFileParsingTesting = list(), list()
    dsspToUse = namedtuple("dsspToUse", ["file", "charChain"])
    proteinNotUse = set()
    with open("./dataset/{}.txt".format(fileName)) as f:
        for line in f:
            proteinNotUse.add(line[:4])
            if len(proteinNotUse) <= trainingNumber:
                resultFileParsingTraining.append(dsspToUse(line[:4], line[4]))
            else:
                resultFileParsingTesting.append(dsspToUse(line[:4], line[4]))
    return resultFileParsingTraining, resultFileParsingTesting


def parserDsspFile(dsspFileName, dsspDirectoryName, charChain):
    """
    Fonction effectuant le parsing du fichier dsspfileName.
    On considère un caractère spécifique(charChain) pour déterminer la séquence d'acides aminés stocké dans "proteinSequence"
    et retirer la structure se rapportant à ces acides aminés pris qui est stocké dans proteinStructure.

    :param dsspFileName: nom du fichier dssp à parser
    :param dsspDirectoryName: nom du dossier dssp où l'on peut trouver le dsspFileName
    :param charChain: caractère spécifique pour déterminer la séquence d'acides aminés
    :return: namedtuple proteinData("proteinIdentifier", "proteinSequence", "proteinStructure")
                                    où proteinIdentifier est le dsspFileName additioné avec le charChain.
    """

    indexBeginSequence = 28
    proteinData = namedtuple("proteinData", ["proteinIdentifier", "proteinSequence", "proteinStructure"])
    dsspFileToUse = "./dataset/{}/{}.dssp".format(dsspDirectoryName, dsspFileName)
    with open(dsspFileToUse) as readFile:
        linesList = readFile.readlines()
        sequenceProtein, structureProtein = getSeqAndStructProtein(linesList[indexBeginSequence:], charChain)
    return proteinData(dsspFileName + charChain, sequenceProtein, structureProtein)


def getSeqAndStructProtein(linesList, charChain):
    """
    Fonction retirant les acides aminés dont la caractère déterminant la séquence est identique à charChain
    et la structure s'y rapportant.

    :param linesList: Liste de tout les lignes contenu dans le fichier dssp à partir de la 1ere ligne des différentes séquences.
    :param charChain: caractère spécifique pour déterminer la séquence d'acides aminés
    :return: sequence: string avec les acides aminés, structure: string avec la structure qui est associé à la sequence.
    """
    structure = ""
    sequence = ""
    charChainIndex = 11
    aminoAcidIndex = 13
    structureSymbolIndex = 16
    for line in linesList:
        if line[charChainIndex] == charChain:
            if line[aminoAcidIndex] not in "BZX":
                # si minuscules répresentent ponts dissulfures donc on remplace par une cystéine C
                if line[aminoAcidIndex].islower():
                    sequence += "C"
                else:
                    sequence += line[aminoAcidIndex]
                structure += symbolsTable[line[structureSymbolIndex].upper()]
    return sequence, structure


def writeInFastaFile(data, dsspDirectoryName, training=False):
    """
    Fonction écrivant les données parser des fichiers dssp dans un fichier .fasta ayant comme structure:
    >identifier
    MTAEPSIVA
    CCEECCCH

    :param training
    :param data: données parser de tous les fichiers dssp
    :param dsspDirectoryName: nom du dossier dssp où l'on peut trouver le dsspFileName
    """
    if training:
        writeFile = open("dataFrom{}Training.fasta".format((dsspDirectoryName[0].upper() + dsspDirectoryName[1:])),
                         "w")
    else:
        writeFile = open("dataFrom{}Testing.fasta".format((dsspDirectoryName[0].upper() + dsspDirectoryName[1:])), "w")
    for item in data:
        writeFile.write(
            ">" + item.proteinIdentifier + "\n" + item.proteinSequence + "\n" + item.proteinStructure + "\n")


def main(dsspDirectoryToUse, cathFileToUse, trainingNumber=3000, test=False):
    """
    Fonction faisant appel aux fonctions parserCathFile et parserDsspFile, qui aboutit à
    la création d'un fichier .fasta contenant les différentes séquences et structure associées.

    :param test: valeur booléen pour différencier cath_info et cath_info_test
    :param trainingNumber
    :param dsspDirectoryToUse: nom du dossier dssp où l'on peut trouver le dsspFileName
    :param cathFileToUse: fichier à parser soit CATH_info ou CATH_info_test
    """
    cathParsingFileTraining, cathParsingFileTesting = parserCathFile(cathFileToUse, trainingNumber)
    dsspParsingFileTraining, dsspParsingFileTesting = list(), list()

    if not test:
        for item in cathParsingFileTraining:
            dsspParsingFileTraining.append(parserDsspFile(item.file, dsspDirectoryToUse, item.charChain))
        writeInFastaFile(dsspParsingFileTraining, dsspDirectoryToUse, training=True)

    for item in cathParsingFileTesting:
        dsspParsingFileTesting.append(parserDsspFile(item.file, dsspDirectoryToUse, item.charChain))
    writeInFastaFile(dsspParsingFileTesting, dsspDirectoryToUse)


if __name__ == '__main__':
    main("dssp_test", "CATH_info_test", trainingNumber=0, test=True)
    main("dssp", "CATH_info")
